------------------------------------------------------------------------------
--                                                                          --
--                            GNAT2WHY COMPONENTS                           --
--                                                                          --
--                   C E _ P R E T T Y _ P R I N T I N G                    --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                     Copyright (C) 2018-2019, AdaCore                     --
--                                                                          --
-- gnat2why is  free  software;  you can redistribute  it and/or  modify it --
-- under terms of the  GNU General Public License as published  by the Free --
-- Software  Foundation;  either version 3,  or (at your option)  any later --
-- version.  gnat2why is distributed  in the hope that  it will be  useful, --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of  MERCHAN- --
-- TABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public --
-- License for  more details.  You should have  received  a copy of the GNU --
-- General  Public License  distributed with  gnat2why;  see file COPYING3. --
-- If not,  go to  http://www.gnu.org/licenses  for a complete  copy of the --
-- license.                                                                 --
--                                                                          --
-- gnat2why is maintained by AdaCore (http://www.adacore.com)               --
--                                                                          --
------------------------------------------------------------------------------

with Ada.Strings.Fixed;        use Ada.Strings.Fixed;
with Ada.Text_IO;
with Ada.Unchecked_Conversion;
with Csets;                    use Csets;
with Gnat2Why.CE_Utils;        use Gnat2Why.CE_Utils;
with Interfaces;               use Interfaces;
with Namet;                    use Namet;
with SPARK_Atree;              use SPARK_Atree;
with SPARK_Util;               use SPARK_Util;
with Uintp;                    use Uintp;

package body Ce_Pretty_Printing is

   function Size (S : String) return Integer is
     (if S (S'First + 1) = 'x' then
           4 * (S'Length - 2)
      else
        (S'Length - 2));
   --  Size returns the associate binary size of a #b or #x number (to help
   --  when building an unsigned integer).

   --  This package is generic so that part of the work done can be shared
   --  between 32bit and 64 bits float numbers.
   generic
      type T_Unsigned is mod <>;
      type T_Float is digits <>;
   package Print_Conversion is

      Bound : constant Integer := T_Unsigned'Size;

      function StringBits_To_Floatrepr (Sign, Significand, Exp : String)
                                        return T_Unsigned;
      --  Transform three stringbits into a single unsigned modular number
      --  (representing a float).

      function Unsigned_To_Float_String (U : T_Unsigned)
                                         return String;
      --  Convert an unsigned number to string representation of a float.

      function StringBits_To_Float_Approx (Sign, Significand, Exp : String)
                                           return String
      is
        (Unsigned_To_Float_String (
           StringBits_To_Floatrepr (Sign, Significand, Exp)));

   end Print_Conversion;

   package body Gen_Print is

      function Beautiful_Source_Name (Ty : Entity_Id) return String
        with Pre => Is_Discrete_Type (Ty);
      --  Does the same as Source_Name except for types defined in Standard
      --  which we print with Upper case letter after each '_'.

      ---------------------------
      -- Beautiful_Source_Name --
      ---------------------------

      function Beautiful_Source_Name (Ty : Entity_Id) return String is
      begin
         if Is_Standard_Entity (Ty) then
            Get_Unqualified_Name_String (Chars (Ty));
            declare
               --  This name is all lower case
               Name     : String := Name_Buffer (1 .. Name_Len);
               --  Decides when the next element should be switch to upper case
               To_Upper : Boolean := True;

            begin
               for I in Name'Range loop
                  if To_Upper then
                     Name (I) := Fold_Upper (Name (I));
                  end if;

                  if Name (I) = '_' then
                     To_Upper := True;
                  else
                     To_Upper := False;
                  end if;
               end loop;
               return Name;
            end;

         else
            return Source_Name (Ty);
         end if;

      end Beautiful_Source_Name;

      --------------------
      -- Print_Discrete --
      --------------------

      function Print_Discrete (Nb : String; Nb_Type : Entity_Id) return String
      is
         Nb_Value : Uint;
      begin
         --  Handle exception from UI_From_String
         begin
            Nb_Value := UI_From_String (Nb);
         exception
            when others => return Nb;
         end;

         --  If one of the bound is not known, we cannot evaluate the type
         --  range so we cannot decide if we alter printing.
         if not Compile_Time_Known_Value (Type_Low_Bound (Nb_Type)) or else
            not Compile_Time_Known_Value (Type_High_Bound (Nb_Type))
         then
            return Nb;
         end if;

         --  Beginning of safe computations.
         declare
            --  Type informations
            Low_Bound  : constant Uint   :=
              Expr_Value (Type_Low_Bound (Nb_Type));
            High_Bound : constant Uint   :=
              Expr_Value (Type_High_Bound (Nb_Type));
            Type_Range : constant Uint   := High_Bound - Low_Bound;
            Type_Name  : constant String := Beautiful_Source_Name (Nb_Type);

            --  Difference calculations
            Diff_To_High : constant Uint := abs (Nb_Value - High_Bound);
            Diff_To_Low  : constant Uint := abs (Nb_Value - Low_Bound);
            Side         : String (1 .. 1);
         begin

            --  If the range of type is too small, we do nothing. If the type
            --  we are given is internal then we don't want to print it as it
            --  would confuse the user.
            --  Example: type Data_T is array (1 .. 1000) of Integer;
            --  There is an internal type Tdata_tD1 for range (1..1000) for
            --  indices: we don't want to print Tdata_tD1'First.
            if Type_Range <= UI_From_Int (Bound_Type) or else
              (not Comes_From_Source (Nb_Type) and then
               not Is_Standard_Entity (Nb_Type))
            then
               return Nb;
            end if;

            --  Nb is closer to the highest bound
            if Diff_To_High <= Diff_To_Low then

               if Diff_To_High = 0 then
                  return Type_Name & "'Last";

               elsif Diff_To_High < Bound_Value then
                  Side := (if Nb_Value < High_Bound then "-" else "+");
                  return Type_Name & "'Last" & Side & UI_Image (Diff_To_High);

               else
                  return Nb;
               end if;

            --  We don't want to print Natural'First + 5 as counterexample. So,
            --  there is a special case when Low_Bound of the type is in
            --  0 .. 1.
            elsif Low_Bound = 0 or else Low_Bound = 1 then
               return Nb;

            else
               if Diff_To_Low = 0 then
                  return Type_Name & "'First";

               elsif Diff_To_Low < Bound_Value then
                  Side := (if Nb_Value < Low_Bound then "-" else "+");
                  return Type_Name & "'First" & Side & UI_Image (Diff_To_Low);

               else
                  return Nb;
               end if;
            end if;
         end;
      end Print_Discrete;

   end Gen_Print;

   --  Start of package body for Print_Conversion
   package body Print_Conversion is

      pragma Assert (T_Unsigned'Size = T_Float'Size);

      function StringBits_To_Unsigned (S : String) return T_Unsigned;
      --  This transforms a number written in bin #b0101 or hex #x5 to an
      --  unsigned integer. (Inside a generic package so the size of unsigned
      --  integer can vary: checks for the size are done outside this
      --  function).

      ----------------------------
      -- StringBits_To_Unsigned --
      ----------------------------

      function StringBits_To_Unsigned (S : String) return T_Unsigned
      is
      begin
         pragma Assert (S (S'First) = '#');
         return T_Unsigned'Value
           (if S (S'First + 1) = 'x' then
              "16#" & S (S'First + 2 .. S'Last) & "#"
            elsif S (S'First + 1) = 'b' then
               "2#" & S (S'First + 2 .. S'Last) & "#"
            else
              raise Program_Error);
      end StringBits_To_Unsigned;

      -----------------------------
      -- StringBits_To_Floatrepr --
      -----------------------------

      function StringBits_To_Floatrepr (Sign, Significand, Exp : String)
                                       return T_Unsigned
      is
         I_Sign           : constant T_Unsigned :=
           StringBits_To_Unsigned (Sign);
         I_Significand    : constant T_Unsigned :=
           StringBits_To_Unsigned (Significand);
         Size_Significand : constant Integer := Size (Significand);
         I_Exp            : constant T_Unsigned :=
           StringBits_To_Unsigned (Exp);
      begin
         return I_Sign * 2 ** (Bound - 1) +
           I_Exp * 2 ** Size_Significand +
           I_Significand;
      end StringBits_To_Floatrepr;

      ------------------------------
      -- Unsigned_To_Float_String --
      ------------------------------

      function Unsigned_To_Float_String (U : T_Unsigned) return String is
         function Convert is new Ada.Unchecked_Conversion
           (Source => T_Unsigned,
            Target => T_Float);
         package F_IO is new Ada.Text_IO.Float_IO (T_Float);
         Result        : T_Float;
         Result_String : String (1 .. Bound);
      begin
         if Convert (U)'Valid then

            --  Unchecked conversion
            Result := Convert (U);

            --  To get nicer results we print small, exactly represented
            --  integers as 123.0 and not in the scientific notation. The
            --  choice of what is "small" is arbitrary; it could be anything
            --  within +/- 2**24 and +/- 2**54 (bounds excluded) for single and
            --  double precision floating point numbers, respectively.

            if abs (Result) < 1000.0
              and then Result = T_Float'Truncation (Result)
            then
               F_IO.Put (To   => Result_String,
                         Item => Result,
                         Aft  => 0,
                         Exp  => 0);
            else
               F_IO.Put (To   => Result_String,
                         Item => Result,
                         --  In the case of long_float, we print 10 decimals
                         --  and we print 7 in case of short float.
                         Aft  => (if Bound = 64 then 10 else 7),
                         Exp  => 1);
            end if;

            return Trim (Source => Result_String,
                         Side   => Ada.Strings.Both);
         else
            return "";
         end if;
      end Unsigned_To_Float_String;

   end Print_Conversion;

   --------------------------
   -- StringBits_To_Approx --
   --------------------------

   function StringBits_To_Approx (Sign, Significand, Exp : String)
                                  return String
   is
   begin
      pragma Assert (Size (Sign) = 1);
      if Size (Significand) <= 23 then
         pragma Assert (Size (Exp) = 8);
         pragma Assert (Size (Significand) = 23);
         declare
            package P is new Print_Conversion (Unsigned_32, Float);
         begin
            return P.StringBits_To_Float_Approx (Sign, Significand, Exp);
         end;
      else
         pragma Assert (Size (Exp) = 11);
         pragma Assert (Size (Significand) = 52);
         declare
            package P is new Print_Conversion (Unsigned_64, Long_Float);
         begin
            return P.StringBits_To_Float_Approx (Sign, Significand, Exp);
         end;

      end if;
   end StringBits_To_Approx;

   -----------------
   -- Print_Fixed --
   -----------------

   function Print_Fixed (Small : Ureal; Nb : String) return String
   is
   begin
      declare
         Nb_Int : constant Uint := UI_From_String (Nb);
         Nb_Real : constant Ureal := UR_From_Uint (Nb_Int) * Small;
      begin
         if Nb_Real = UR_From_Uint (UR_Trunc (Nb_Real)) then
            return UI_Image (UR_To_Uint (Nb_Real), Decimal);
         else
            declare
               Num_Small : constant String :=
                 UI_Image (Norm_Num (Small), Decimal);
               Den_Small : constant String :=
                 UI_Image (Norm_Den (Small), Decimal);
            begin
               return Nb
                 & (if Num_Small /= "1" then "*" & Num_Small else "")
                 & (if Den_Small /= "1" then "/" & Den_Small else "");
            end;
         end if;
      end;
   exception when others =>
         return "";
   end Print_Fixed;

end Ce_Pretty_Printing;
